require 'singleton'
require 'framework/mixins/notifier'
require 'socket'
require 'monitor'

module Mondrian
	
	Breakpoint = Struct.new(:file_path, :line)
	# probably need to add field for debugger_line,
	# representative of state when current debug process
	# was first run
	class Breakpoint
		def inspect
			"#{file_path}:#{line+1}"
		end
	end

	class Debugger
		include Singleton
		include Kinshasa::Notifier
		include MonitorMixin
		
		event :breakpoint_added, :breakpoint_removed, :stopped, :resumed, :disconnected
		attr_reader :breakpoints
		
		LOGGING = false
		
		def log(msg)
			puts msg if (LOGGING)
		end			
			
		def initialize
			@breakpoints = []
			ExecutionManager.instance.subscribe_all(self, method(:on_execution_update))
			@state = :disconnected
			@expecting = false
		end
		
		def on_execution_update(sender, message, data)
			if message == :execution_begin
				connect
			elsif message == :exection_end
				disconnect
			end			
		end
		
		def add_breakpoint(bp)
			return if @breakpoints.include? bp
			@breakpoints << bp
			notify_breakpoint_added(bp)
			send_breakpoint(bp)
		end
		
		def remove_breakpoint(bp)
			if @breakpoints.delete(bp)
				notify_breakpoint_removed(bp)
				send_command("del")				
				send_all_breakpoints
			end
		end
		
		def toggle_breakpoint(bp)			
			@breakpoints.include?(bp) ? remove_breakpoint(bp) : add_breakpoint(bp)
		end
		
		# check usages of this -- what if the user doesn't save?
		def replace_breakpoint(old, new)
			remove_breakpoint(old)
			add_breakpoint(new)
		end		
		
		def continue
			unstop
			send_command_r("cont")
		end
		
		def step_over
			unstop
			send_command_r("next")
		end
		
		def step_in
			unstop
			send_command_r("step")
		end
		
		def step_out
			unstop
			send_command_r("finish")
		end        
    
    def eval(expr, &block)			
			@expecting = true
			@callback = block			
			send_command_r("eval " + expr)
		end

		def connected?
			[:running, :stopped].include? @state
		end

		def stopped?
			@state == :stopped
		end

		##########
		private 
		##########
		
		def unstop
			@state = :running
			notify_resumed
		end	
		
		def send_breakpoint(bp)			
			send_command("break #{bp.file_path}:#{bp.line+1}")
		end
		
		def send_all_breakpoints
			@breakpoints.each { |bp| send_breakpoint(bp) }
		end

		def connect
			unless ExecutionManager.instance.debugging?
				@rsocket = @socket = nil
				return			
			end
			return if connected?
			
			Thread.new do
				10.times do 
					log "TRYING TO CONNECT #{@socket}, #{@rsocket}"
					#synchronize do
						begin						
							@socket ||= TCPSocket.new('localhost', 8990)
							@rsocket ||= TCPSocket.new('localhost', 8989)
						rescue				
							@socket = @rsocket = nil
							log "FAILED TO CONNECT"
						end
					#end
					unless @socket.nil?
						@state = :running
						send_all_breakpoints
						listen_for_output(@socket, "s")
						listen_for_output(@rsocket, "r")           
						log "CONNECTED!"
            break
					else
            sleep(1)
          end
				end
			end
		end
		
			
		def disconnect
			#synchronize do
				log "DISCONNECTING #{@socket}, #{@rsocket}"
				was = connected?
				@socket.close if @socket
				@socket = nil
				@rsocket.close if @rsocket
				@rsocket = nil
				@state = :disconnected
				notify_disconnected if was
			#end			
		end	
			
		def send_command_r(cmd)
      send_command(cmd, @rsocket, "r")
    end
    
    def send_command(cmd, sock = nil, msg = "s")			
			return unless connected?
			sock ||= @socket
			begin
				log "ABOUT TO SEND: #{msg}: " + cmd
				sock.puts cmd
				log "SENT: #{msg}:"  + cmd 
			rescue Exception
				log "ERROR SENDING #{msg}: #{$!}"
				disconnect
			end
		end
		
		def handle_stop(file, line)
		  @state = :stopped
			SessionManager.instance.chore {
				notify_stopped(Breakpoint.new(file, line))
			}
		end			

		def listen_for_output(sock, msg)
			Thread.new do								
				begin
				while line = sock.gets
					log "OUTPUT #{msg}: " + line
					if @expecting
						@callback.call(line)
						@expecting = false
						@callback = nil
					elsif line =~ /^(.*?):(\d*): /
						log "STOP #{msg}: #{$1}, #{$2}"
						handle_stop($1.clone, $2.to_i)
					end					
				end
				rescue
					log "ERROR in OUTPUT THREAD #{msg}: #{$!} : #{$!.backtrace.join("\n")}"
				end				
				log "OUTPUT THREAD FINISHED #{msg}"
				disconnect
			end
		end
	end

	at_exit do
		begin
			Debugger.instance.disconnect
		rescue Exception
		end
	end	
end