# tokenencoder.rb
#
# Contributors:: Oliver Smith


module Kinshasa

    # Utilities for turning code into an encoded token string.
    #
    # For a string of tokens, an encoded token string can be constructed
    # which is exactly the same length as the original string and stores
    # information about token types.
    #
    # Each token type, like :string or :operator, is given a unique
    # character from +TokenCharacters+ which represents it.  An "encoded"
    # token is that token's character followed by as many 
    # +TokenFillCharacter+s it takes to equal the length of the token,
    # e.g. 'a~~~~'.  A string of encoded tokens can sit beside the original
    # code string to facilitate easy manipulation.

    class TokenEncoder
        TokenCharacters = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_' unless defined? TokenCharacters
        TokenFillCharacter = '~' unless defined? TokenFillCharacter
        TokenNullCharacter = '.' unless defined? TokenNullCharacter
        
        def initialize(symbols, groups = {}) # FIXME: should a TokenEncoder just be seeded with a Lexer?
            @tokens = {}
            @groups = {}
        
            symbols.each { |symbol|
                add_token symbol
            }
            
            groups.each { |group, tokens|
                add_group group, tokens
            }
        end
        
        
        # Returns a regexp for the token +symbol+, or for any token.
        
        def match(symbol = nil)
            return match_any unless symbol
            
            send(symbol)
        end


        # Returns a regexp which matches any encoded token.

        def match_any
            /\w#{TokenFillCharacter}*/o
        end
        
        
        # Returns the token symbol for +encoded_token+.
        
        def token_symbol(encoded_token)
            @tokens.invert[encoded_token[0].chr]
        end


        # Returns an encoded token string for token +symbol+.

        def encode_token(symbol, length)
            @tokens[symbol] + TokenFillCharacter * (length - 1)
        end
        
        
        # Returns whether +encoded_token+ is of type +symbol+.
        
        def encoded_token_is?(encoded_token, symbol) # FIXME: doesn't work for groups
            token_symbol(encoded_token) == symbol
        end


        # Returns the number of tokens in +encoded_tokens+.

        def count_tokens(encoded_tokens)
            encoded_tokens.count(TokenCharacters)
        end
        
        
        # Returns a padded null (non-token) string of +length+ characters.
        
        def null_pad(length)
            TokenNullCharacter * length
        end
        
        
        # Assigns a token a unique character and adds utility methods for it.
        
        def add_token(symbol)
            @tokens[symbol] = TokenCharacters[@tokens.size].chr
                
            add_token_methods(symbol)
        end
        
        
        # Creates a new group +symbol+ which matches anything in +tokens+.
        
        def add_group(symbol, tokens)
            @groups[symbol] = tokens
            
            add_group_methods(symbol, tokens)
        end
        
        
        # Adds utility methods for token +symbol+.
        #
        # e.g. for token :whitespace, creates the methods +whitespace+ and
        # +not_whitespace+, which return matching regexps.
        
        protected
        def add_token_methods(symbol)
            eval <<-fin
                def #{symbol}
                    /#{@tokens[symbol]}#{TokenFillCharacter}*/o
                end
                
                
                def not_#{symbol}
                    /[^#{@tokens[symbol]}#{TokenFillCharacter}#{TokenNullCharacter}]#{TokenFillCharacter}*/o
                end
            fin
        end


        # Adds utility methods for token group +symbol+.
        #
        # A group allows for easy matching of any token in a list.
        
        protected
        def add_group_methods(symbol, tokens)
            eval <<-fin
                def #{symbol}
                    /[#{tokens.map { |token| @tokens[token] }.to_s}]#{TokenFillCharacter}*/o
                end
                
                
                def not_#{symbol}
                    /[^#{tokens.map { |token| @tokens[token] }.to_s}#{TokenFillCharacter}#{TokenNullCharacter}]#{TokenFillCharacter}*/o
                end
            fin
        end

    end

end    


