# A class for storing objects in a tree representation.
#
# Contributors:: Oliver Smith


require 'framework/mixins/notifier'


module Kinshasa

    # A class representing a node in a +TreeList+.  Note that the events
    # advertised are only triggered in the root node, which is a +TreeList+.
    # 
    # FIXME: need to find a way to avoid moves that break the structure

    class TreeNode
        include Notifier
    
        attr_reader :root, :parent, :before, :after, :first
        attr_accessor :data
        
        event :node_added, :node_moved, :node_deleted
        
        def initialize(_data = nil)
            @root = @parent = @before = @after = @first = nil

            self.data = _data
        end
        
        
        def root?
            self == @root
        end
        
        
        def data=(_data)
            @data = _data
        end
        

        def last
            node = @first
            node = node.after while node and node.after
            node
        end
        
        
        def move_after(node)
            assert_node(node)
            
            remove
            insert(node.parent, node, node.after)
            
            @root.notify_node_moved(self)
        end
        
        
        def move_before(node)
            assert_node(node)
            
            remove
            insert(node.parent, node.before, node)

            @root.notify_node_moved(self)
        end
        
        
        def move_under(node)
            assert_node(node)

            remove
            insert(node.parent, nil, nil)
            
            @root.notify_node_moved(self)
        end


        def insert_after(node)
            assert_node(node)
            
            node.insert(parent, self, after)
            
            @root.notify_node_added(node)

            node
        end
        
        
        def insert_before(node)
            assert_node(node)

            node.insert(parent, before, self)
            
            @root.notify_node_added(node)

            node
        end
        
        
        def append(node)
            assert_node(node)

            node.insert(self, last, nil)
            
            @root.notify_node_added(node)

            node
        end
        
        
        def children
            children = []
            
            node = first
            
            while node
                children << node 
                node = node.after
            end
            
            children
        end
        
        
        def each_child
            children.each { |node|
                yield node
            }
        end


        # A default implementation of +to_s+ which returns a simple string
        # representation of the node and its children's +data+.
        
        def to_s(level = 0)
            str = "\t" * level + "- " + data.to_s + "\n"

            level += 1
            
            each_child { |node|
                str += node.to_s(level)
            }
            
            str
        end
        
        
        protected
        def before=(node)
            @before = node
        end
        
        
        protected
        def after=(node)
            @after = node
        end
        
        
        protected
        def root=(node)
            @root = node
        end
        
        
        protected
        def first=(node)
            @first = node
        end
        
        
        protected
        def parent=(node)
            @parent = node
        end
                

        protected
        def remove
            old_before = before
            old_after = after
            
            old_before.after = old_after if old_before
            old_after.before = old_before if old_after
        end
        
        
        protected
        def insert(parent, before, after)
            parent.first = self if !parent.first or !before
            
            before.after = self if before
            after.before = self if after
        
            self.root = parent.root
            self.parent = parent
            self.before = before
            self.after = after
        end
        
        
        private                
        def assert_node(*nodes)
            nodes.each { |node|
                raise ArgumentError unless node.kind_of?(TreeNode)
            }
        end
        
    end
    


    # A class for representing objects in a tree.  +TreeList+ is itself a
    # +TreeNode+, and acts as the root node.  +TreeList+s are nestable, so that
    # a +TreeList+ root node may be added as a child of a +TreeNode+.

    class TreeList < TreeNode
        # Creates a new +TreeList+.
        
        def initialize(_data = nil)
            super(_data)

            @root = self
        end

    end

end