# file.rb
#
# Contributors:: Oliver Smith


require 'framework/common/utilities'
require 'framework/datasource'
require 'framework/mixins/editable'
require 'framework/resource'
require 'framework/tokenstring'


module Kinshasa

    # A class for representing a filetype.  Default behavior is to handle any
    # file with its extension in +extensions+ or a hash-bang style program name
    # in +programs+.

    class FileTypeResource < Resource
        attr_reader :file_class, :extensions, :programs
    
        def initialize(file_class, extensions = [], programs = [], &handles_proc)
            @extensions = extensions
            @programs = programs
            
            super(file_class, [ String ], &handles_proc)
        end
        
        
        def add_extensions(extensions)
            extensions.each { |e|
                @extensions << e.downcase
            }
        end
    
    
        def add_programs(programs)
            programs.each { |p|
                @programs << p
            }
        end


        def handles?(path)
            return false unless super
        
            return false if File.directory?(path)
            return false unless File.exists?(path)
            
            if not(extensions.empty? or File.extname(path) == '')
                return extensions.include?(File.extname(path).downcase)
            elsif not programs.empty?
                return programs.include?(Common::get_hash_bang_program(path))
            end
            
            true
        end

    end



    # A class for representing a generic file.  Subclass for specific 
    # filetypes.

    class GenericFile < DataSource
        include Editable
    
        attr_reader :path
        
        FileTypeResource.new(GenericFile)
        
        
        # Creates a FileTypeResource for the new subclass.
        
        def GenericFile.inherited(klass)
            FileTypeResource.new(klass)
        end
        
        
        # Returns the FileTypeResource for this class.
        
        def self.filetype
            FileTypeResource.resources.each { |r|
                return r if r.data == self
            }
        end
        
        
        # Registers one or more extensions for this filetype
        
        def self.extension(*extensions)
            filetype.add_extensions extensions
        end
        
        
        # Registers one or more hash-bang style programs for this filetype.

        def self.program(*programs)
            filetype.add_programs programs
        end

        
        # Creates a new file at +path+.  Children should override for more
        # appropriate behavior if necessary.
        
        def self.create(path)
            raise Errno::EEXIST if File.exists?(path)
            
            File.new(path, 'w').close
            
            self.new(path)
        end
        
        
        # Where +path+ points to an existing file.
        
        def initialize(path)
            raise Errno::ENOENT if File.directory?(path)
            raise Errno::ENOENT unless File.exists?(path)
        
            @path = File.expand_path(path)
            
            super(File.basename(@path))
        end
        
        
        # Saves this file.  Override for specific behavior.
        
        def save
            clean
        end
        
        
        # Closes the file.  Override for specific behavior.
        
        def close
        end
        
    end
    
    
    
    # A GenericFile which holds a tokenstring.
    
    class TextFile < GenericFile
        attr_reader :tokenstring
        
        def tokenstring
            unless defined? @tokenstring
                text = ''
            
                File.open(path, 'r') { |f|
                    text = f.read
                }
                
                @tokenstring = token_string_class.new(text)
                
                subscribe_to_events
            end
            
            @tokenstring
        end
        
        
        def save
            File.open(path, 'w') { |f|
                f.write(@tokenstring.text)
            }
            
            super
        end
        
        
        def close
            unsubscribe_from_events
        end
        

        private

        def token_string_class
            TokenStringResource.resource_for(self)
        end
        

        private
        
        def subscribe_to_events
            @tokenstring.subscribe_text_lexed(self) { dirty }
        end
        
        
        private
        
        def unsubscribe_from_events
            @tokenstring.unsubscribe_all(self)
        end
        
    end

end