# folder.rb
#
# Contributors:: Oliver Smith


require 'framework/common/treelist'
require 'framework/common/utilities'
require 'framework/datasource'
require 'framework/file'


module Kinshasa

    # A class for managing a folder and its files.

    class Folder < DataSource
        attr_reader :name, :path, :tree
        
        # Creates a new folder at +path+.
        
        def self.create(path)
            raise Errno::EEXIST if File.exists?(path)
            
            Dir.mkdir(path) unless File.exists?(path)
            
            self.new(path)
        end
        
        
        # Where +path+ is a folder.
        
        def initialize(path)
            raise Errno::ENOENT unless File.directory?(path)
        
            @path = File.expand_path(path)
            
            super(File.basename(@path))
                        
            build_tree
        end
        
        
        # Returns whether +abs_path+ is a non-linked folder under this folder.

        def contains_folder?(abs_path)
            raise ArgumentError, "Path (#{abs_path}) is not absolute!" unless Common::is_absolute? abs_path

            return Common::is_real_directory?(abs_path) && Common::is_subpath?(abs_path, @path)
        end
        
        
        # Returns whether +abs_path+ is a file contained under this folder.

        def contains_file?(abs_path)
            raise ArgumentError, "Path (#{abs_path}) is not absolute!" unless Common::is_absolute? abs_path
        
            return File.file?(abs_path) && Common::is_subpath?(abs_path, @path)
        end

        
        # Returns the absolute names of subfolders non-recursively.
        
        def folder_paths(abs_path = @path)
            raise ArgumentError, "Path (#{abs_path}) does not refer to a folder beneath this folder!" unless contains_folder?(abs_path)

            folders = Dir.new(abs_path).entries.collect { |entry| File.join(abs_path, entry) }.select { |entry| contains_folder? entry }
            
            folders.sort { |a, b| File.basename(a).downcase <=> File.basename(b).downcase }
        end
        

        # Returns the absolute names of files non-recursively.
        
        def file_paths(abs_path = @path)
            raise ArgumentError, "Path (#{abs_path}) does not refer to a file beneath this folder!" unless contains_folder?(abs_path)
        
            files = Dir.new(abs_path).entries.collect { |entry| File.join(abs_path, entry) }.select { |entry| contains_file? entry }
            
            files.sort { |a, b| File.basename(a).downcase <=> File.basename(b).downcase }
        end


        # Loads the contained files into a TreeList, instantiating file or
        # folder objects for each.

        private
        def build_tree
            @tree = TreeList.new(self)

            folder_paths(@path).each { |folder_path|
                @tree.append(Folder.new(folder_path).tree)
            }
            
            file_paths(@path).each { |file_path|
                @tree.append(TreeNode.new(FileTypeResource.resource_for(file_path).new(file_path)))
            }
        end
        
    end
    
end