# lexer.rb
#
# Contributors:: Oliver Smith


module Kinshasa

    # A mixin which provides +Lexer+ with its +token+ method.
   
    module LexerInternals

        def self.append_features(receiver)
            super
            
            receiver.extend(ClassMethods)
        end
    
        
        # Returns a list of the +Lexer+'s token symbols. 
        
        def tokens
            self.class.tokens
        end
        
        
        # Returns a hash of the +Lexer+'s groups.
        
        def groups
            self.class.groups
        end
 

        module ClassMethods

            # Returns a list of the +Lexer+'s token symbols. 

            def tokens
                (defined? @tokens) ? @tokens : []
            end
            
            
            # Returns a hash of the +Lexer+'s groups.
            
            def groups
                (defined? @groups) ? @groups : {}
            end
            
            
            # Adds one or more token symbols.
        
            private
            def token(*symbols)
                @tokens = [] unless defined? @tokens
            
                symbols.each { |symbol|
                    @tokens << symbol
                }
            end
            
            
            # Creates a symbol to refer to a group of tokens.
            
            private
            def group(symbol, tokens)
                @groups = {} unless defined? @groups
                
                @groups[symbol] = tokens
            end

        end    

    end



    # A base class for token lexers to inherit.

    class Lexer
        include LexerInternals            
    
        # Turns a +keywords+ list into a fast regexp which matches any of those keywords.
        #
        # A keyword is assumed to end at a word boundary, e.g.
        # [ 'if', 'then', 'else' ] becomes /\A(if|then|else)\b/.
    
        def keyword_regexp(keywords)
            Regexp.new('\A(' + keywords.map { |keyword| Regexp.escape(keyword) }.join('|') + ')\b')
        end


        # Turns an +operators+ list into a fast regexp which matches any of those operators.
        #
        # e.g. [ '+', '-', '*' ] becomes /\A()/.

        def operator_regexp(operators)
            Regexp.new('\A(' + operators.map { |operator| Regexp.escape(operator) }.join('|') + ')')
        end
        
        
        # Children should return a tuple containing the next token in +code+.
        #
        # e.g. for a string of ruby code "raise 'Base class method called!'",
        # a ruby +Lexer+ would return ":keyword, 'raise'".
        #
        # +code+ is not a string, but rather a +StringScanner+ whose position
        # is set to the start of the unlexed code.  +code+ should be updated
        # by this call to point to the position following the lexed token
        # (this is usually done for us by the +StringScanner::scan+ call).
        #
        # +context+ is a +LexerContext+ which provides information about
        # the current lex position.
        
        def next_token(code, context)
            raise 'Base class method called!'
        end
        
        
        # Returns whether +symbol+ is a group.
        
        def is_group?(symbol)
            return groups.include?(symbol)
        end
        
        
        # Returns whether +symbol+ is a token.
        
        def is_token?(symbol)
            return tokens.include?(symbol)
        end


        # Returns whether +token+ is a member of +group+.
        
        def in_group?(token, group)
            groups[group].include?(token)
        end
    end
    

    
    # Stores information about the current state while lexing.
    #
    # A +LexerContext+ keeps track of things like the last token lexed, or anything
    # the +Lexer+ might need to remember between parsing one token and the next.
    # For instance, a keyword may only be a keyword if it follows an equal sign.
    # The +LexerContext+ could keep a boolean remembering whether the previous non-
    # whitespace token was in fact an equal sign.
    
    class LexerContext
        attr_accessor :last, :last_token # FIXME: last and last_token should be renamed
     
        # Accepts the last token lexed and its token symbol.
        
        def initialize(last, last_token)
            @last = last
            @last_token = last_token
        end
        
    
        # Accepts the last token lexed and its token symbol.
        #
        # Children should return true to force continued lexing even if all changed
        # text has been lexed and the previous token was unchanged.

        def update!(last, last_token)
            @last = last
            @last_token = last_token
            
            false
        end

    end

end
