# resource.rb
#
# Contributors:: Oliver Smith


module Kinshasa

    class ResourceNotFoundError < Exception
    end

    
    
    class ResourceHook
        def initialize(resource_class, callable)
            @resource_class = resource_class
            @callable = callable
        end
        
        
        def resource_for(object)
            @callable.call(@resource_class.resource_for(object))
        end
    end



    # A class for representing a resource, such as an icon, which may be
    # associated with a particular instance, such as the specific icon used
    # to represent files of a certain type.  Subclass this and give it whatever 
    # attributes it needs to represent your resource.  Simply instantiating the 
    # subclass will register the instance under the appropriate type.  The base 
    # class holds a list of all registered +Resource+ types and instances in 
    # +@resources+.
    
    class Resource
        attr_reader :data
    
        @@resources = {}
        
        def Resource.inherited(klass)
            @@resources[klass] = []
        end
        
        
        # Returns the list of +Resource+ objects for this class.
        
        def self.resources
            @@resources[self]
        end
    
    
        # Returns the appropriate resource for +object+.  The appropriate
        # resource is the last one to have registered which signals capability
        # of handling +object+, and is the object returned by the +Resource+'s
        # +data+ method.
        
        def self.resource_for(object)
            resource = nil
        
            if @@resources.include? self
                @@resources[self].each { |r|
                    resource = r if r.handles? object
                }
            end
            
            raise ResourceNotFoundError.new unless resource

            resource.data
        end
        
        
        # Adds a converter hook to this +Resource+ type.  For instance, an 
        # +Icon+ resource can add a hook called 'fox16', which would be used 
        # like: +Icon.fox16.resource_for(object_that_needs_an_icon)+.
        
        def self.hook(hook_name, &block)
            raise NotImplementedError if self.name == 'Resource'
            
            @hooks = {} unless defined? @hooks
            @hooks[hook_name] = block

            eval <<-fin
                def self.#{hook_name}
                    ResourceHook.new(self, @hooks['#{hook_name}'])
                end
            fin
        end
        
        
        # +classes+ is the list of classes this +Resource+ applies to.  
        # +classes+ may be omitted if +handles_proc+ is provided to evaluate
        # whether an object is handled by this instance or not, or both may
        # be used to narrow down the class and fine detail.
        
        def initialize(data = nil, classes = [], &handles_proc)
            @data = data
            @classes = classes
            @handles_proc = handles_proc
            
            @@resources[self.class] << self
        end
        
        
        # Returns +true+ if this Resource instance is associated with +object+.
        # Children can subclass for more fine-tuned behavior, but the default
        # can check whether +object+'s class is found in +@classes+, and if 
        # the +@handles_proc+ passed in to +initialize+ evaluates to +true+.
        
        def handles?(object)
            if @handles_proc
                return (@classes.empty? || @classes.include?(object.class)) && @handles_proc.call(object)
            else
                return @classes.include?(object.class)
            end
        end
    
    end
    
end