# rubylexer.rb
#
# Contributors:: Oliver Smith

require 'framework/lexer'


module Kinshasa

    # A +Lexer+ for the ruby language.

    class RubyLexer < Lexer
        token :indentation, :keyword, :operator, :string, :system_variable, 
              :comment, :number, :identifier, :class_name, :method_name,
              :module_name, :system_command, :regexp, :array_literal, 
              :linebreak, :junk

        group :whitespace, [ :indentation, :linebreak, :junk ]

        
        # Pre-creates a bunch of useful regexps.
        #
        # FIXME: idea - evaluations in regexps are slow, so for delimiters below we could pre-create them all
        
        def initialize
            system_variables = [ 
                '$!', '$@', '$&', '$+', '$`', '$\'', '$=', '$1', '$2', '$3',
                '$4', '$5', '$6', '$7', '$8', '$9', '$~', '$/', '$-0', '$\\',
                '$,', '$.', '$;', '$<', '$>', '$-F', '$0', '$*', '$"', '$$', 
                '$?', '$:', '$-a', '$-d', '$-i', '$-I', '$-K', '$-l', '$-p',
                '$-v', '$-w' 
            ]
        
            operators = [ 
                '[', ']', '**', '!', '++', '--', '^=', '%=', '%', '<<', '>>', 
                '<=>', '<=', '<', '>=', '>', '===', '==', '!=', '=~', '!~',
                '&&', '||=', '||', '...', '..', '?', '{', '}', '/=', '-=', 
                '+=', '|=', '&=', '>>=', '<<=', '*=', '&&=', '**=', '.', '::', 
                ':', '(', ')', ',', '=', '+', '-', '&', '|', '^', '*', '/', '~' 
            ]
        
            keywords = [ 
                '__LINE__', '__FILE__', 'and', 'def', 'end', 'in', 'or', 'self', 
                'unless', 'begin', 'defined?', 'ensure', 'module', 'redo', 
                'super', 'until', 'BEGIN', 'break', 'do', 'false', 'next', 
                'rescue', 'then', 'when', 'END', 'case', 'else', 'for', 'nil', 
                'retry', 'true', 'while', 'alias', 'class', 'elsif', 'if', 'not',
                'return', 'undef', 'yield', 'raise' 
            ] 
        
            operator_methods = [ 
                '[]=', '[]', '**', '!', '~', '+@', '-@', '*', '/', '%', '+', '-', 
                '<<', '>>', '&', '^', '|', '<=>', '<=', '<', '>=', '>', '===', 
                '==', '=~' 
            ]

            @r_system_variables = operator_regexp(system_variables)
            @r_operators        = operator_regexp(operators)
            @r_operator_methods = operator_regexp(operator_methods)
            
            @r_keywords = keyword_regexp(keywords)
            
            @r_linebreak         = /\A(\r\n|[\r\n;])/
            @r_indentation       = /\A[\ \t\f\\]+/
            @r_comment           = /\A#[^\r\n]*/
            @r_number            = /\A\d\w*([.]\d+)?/
            @r_method_name       = /\A[A-Za-z_][\w.]*[?!=]?/
            @r_class_name        = /\A(<<\s*|[A-Z])\w*/
            @r_module_name       = /\A[A-Z]\w*/
            @r_string            = /\A("|'|`)(\\(?:[^\r\n]|\Z)|(?!\\|\1)[^\r\n])*(\1|(?=\r)|$)/
            @r_regexp            = /\A\/(\\(?:[^\r\n]|\Z)|(?!\\|\/)[^\r\n])*(\/[iomx]*|)/
            @r_identifier        = /\A([$]|@|@@)?[A-Za-z_]\w*/
            @r_percent_literal   = /\A%(q|Q|r|w|x)?([^A-Za-z0-9\r\n])/
            @r_multiline_comment = /\A=begin\s.*?(^=end|\z)/m
            @r_junk              = /./
        end


        # Reads the next token tuple from +code+.
        #
        # +code+ is a +StringScanner+ and +context+ is a +RubyContext+.
        #
        # See the documentation for +Lexer::next_token+.

        def next_token(code, context)
            # indentation (all whitespace other than a linebreak)
            return :indentation, code.matched if code.scan @r_indentation
            
            # linebreak
            return :linebreak, code.matched if code.scan @r_linebreak
        
            # single-line comments
            return :comment, code.matched if code.scan @r_comment
            
            # numbers
            return :number, code.matched if code.scan @r_number
        
            # definition acceptor, including overloadable operators
            if context.last_black_token == :keyword
                case context.last_black
                    when "class"
                        return :class_name, code.matched if code.scan @r_class_name
                    when "def"
                        return :method_name, code.matched if code.scan @r_operator_methods
                        return :method_name, code.matched if code.scan @r_method_name
                    when "module"
                        return :module_name, code.matched if code.scan @r_module_name
                end
            end
        
            # keywords
            unless context.last_token == :operator and context.last == '.'
                return :keyword, code.matched if code.scan @r_keywords
            end
        
            # basic strings and system commands
            if code.scan @r_string
                case code[1]
                    when '`'
                        return :system_command, code.matched
                    else
                        return :string, code.matched
                end
            end

            # find some more information about the previous token
            expecting_literal = (context.last_black_token == nil or
                (context.last_black_token == :operator and context.last_black != ')' and context.last_black != ']') or
                context.last_black_token == :keyword or
                context.last_black_token == :comment)
                
            # basic regexps
            if expecting_literal and code.scan @r_regexp
                return :regexp, code.matched
            end
                            
            # % literals
            if expecting_literal and code.match? @r_percent_literal
                literal_type = code[1]
                delimiter    = code[2]
                    
                case delimiter
                    when '('
                        delimiter = ')'
                    when '{'
                        delimiter = '}'
                    when '['
                        delimiter = ']'
                    when '<'
                        delimiter = '>'
                end
                    
                if not (literal_type == 'w' and delimiter =~ /\s/)
                    delimiter = Regexp.escape(delimiter)
    
                    if literal_type == 'r' # add the option characters on regexps
                        if delimiter == "\\\\" # can't allow escaping on backslash delimiters
                            code.scan(/\A%(r)?\\.*?(\\[iomx]*|$)/)
                        else
                            code.scan(/\A%(r)?.(?:\\(?:.|\Z)|(?!\\|#{delimiter}).)*(#{delimiter}[iomx]*|$)/)
                        end
                    else
                        if delimiter == "\\\\" # can't allow escaping on backslash delimiters
                            code.scan(/\A%(q|Q|w|x)?\\.*?(\\|$)/)
                        else
                            code.scan(/\A%(q|Q|w|x)?.(?:\\(?:.|\Z)|(?!\\|#{delimiter}).)*(#{delimiter}|$)/)
                        end
                    end
                        
                    case code[1]
                        when 'w'
                            return :array_literal, code.matched
                            
                        when 'r'
                            return :regexp, code.matched
                            
                        when 'x'
                            return :system_command, code.matched
                            
                        else
                            return :string, code.matched
                    end
                end
            end
                
            # identifiers
            return :identifier, code.matched if code.scan @r_identifier
            
            # multiline comments
            if code.bol? and code.scan @r_multiline_comment
                return :comment, code.matched 
            end
        
            # here documents
            if code.match?(/\A<</) and
                (
                   code.scan(/\A<<("|')(?-m:(.+?))(\1)\s*$.*?(^\2\s*$|\z)/m) or
                   code.scan(/\A<<-("|')(?-m:(.+?))(\1)\s*$.*?(^\s*\2\s*$|\z)/m) or
                   code.scan(/\A<<(\w+)\s*$.*?(^\1\s*$|\z)/m) or
                   code.scan(/\A<<-(\w+)\s*$.*?(^\s*\1\s*$|\z)/m)
                )
                return :string, code.matched
            end
        
            # operators
            return :operator, code.matched if code.scan @r_operators
            
            # system variables
            return :system_variable, code.matched if code.scan @r_system_variables
            
            # the catch-all...accepts any leftover characters
            return :junk, code.matched if code.scan @r_junk
        end

    end

    
    
    # A +LexerContext+ which also saves the last non-whitespace token.
    
    class RubyContext < LexerContext
        attr_accessor :last_black, :last_black_token # FIXME: need better naming convention here
    
        def initialize(last = '', last_token = nil, last_black = '', last_black_token = nil)
            @last_black = last_black
            @last_black_token = last_black_token
            
            update!(last, last_token)
        end
    
    
        # Stores the token if it's non-whitespace.
        #
        # Returns true if the new token is whitespace, meaning that we need
        # to continue lexing even if all changed text has been covered to
        # make sure the next non-whitespace token hasn't changed because of
        # a changed previous one.
        
        def update!(token_text, token_type)
            super
            
            # FIXME: should we give a LexerContext a pointer to the Lexer?  repeating info from RubyString here...
            unless token_type == :indentation || token_type == :linebreak || token_type == :junk
                @last_black       = token_text 
                @last_black_token = token_type
            end
            
            return @last_token != @last_black_token
        end

    end

end
